from inspect import Parameter
import os
import sys
import numpy as np
from time import time
from glob import glob
import tensorflow as tf
from keras import backend as K 
import tensorflow.keras.layers as tflayers
from tensorflow.keras.models import Sequential, load_model
from tensorflow.keras.preprocessing.image import load_img, img_to_array

parameters = sys.argv[1:]

if parameters[0] == 'train':
    train_or_test, kernel_type, topology, model_loss, model_optimizer, model_file_name, data_folder = parameters

    if model_loss == 'cross':
        model_regularizer = None
    elif model_loss == 'cross-l1':
        model_regularizer = 'l1'
    elif model_loss == 'cross-l2':
        model_regularizer = 'l2'

else:
    train_or_test, model_file_name, data_folder = parameters

# train_or_test = 'train'
# kernel_type = '5diamond'
# model_loss = 'cross-l2'
# model_optimizer = 'adam'
# model_file_name = 'suits_model'
# topology = '2'
# data_folder = 'data'

data_folder = os.path.abspath(data_folder)

#Tweakable Parameters
#-------------------------------------
batch_size = 32
epochs = 10
#-------------------------------------
if not model_file_name.endswith('.h5'):
    model_file_name+='.h5'





def diamond_kernel(shape, dtype):
    receptive_field_size = shape[0]*shape[1]
    fan_in = shape[-2] * receptive_field_size
    fan_out = shape[-1] * receptive_field_size
    limit = (6 / (fan_in + fan_out))**0.5
    kernel = tf.random.uniform(shape, minval= -limit, maxval=limit)
    diamond = tf.constant([[0, 0, 1, 0, 0],
                             [0, 1, 1, 1, 0],
                             [1, 1, 1, 1, 1],
                             [0, 1, 1, 1, 0],
                             [0, 0, 1, 0, 0]], dtype = 'float32')
    kernel = tf.multiply(tf.expand_dims(tf.stack((diamond, )*shape[-1], axis = 2), axis = 2), kernel)
    return kernel

class customConv2d(tflayers.Layer):
    def __init__(self, filters = None, filter_size = 5, kernel_type = '5square', model_regularizer = None,  **kwargs):
        self.filters = filters
        self.kernel_type = kernel_type
        self.kernel_size = [filter_size,filter_size]
        self.model_regularizer = model_regularizer
        super(customConv2d, self).__init__(**kwargs)
            
    def build(self, input_shape):
        shape = list(self.kernel_size) + [input_shape[-1], self.filters]
        self.kernel = self.add_weight(
            name = "kernel",
            shape=shape,
            initializer = 'glorot_uniform' if self.kernel_type == '5square' else diamond_kernel,
            regularizer = self.model_regularizer,
            trainable = True
        )
        
        weights = self.kernel.numpy()
        super(customConv2d, self).build(input_shape)
        return self.kernel

    def call(self, inputs):
        return K.conv2d(inputs, self.kernel)
    
    def get_config(self):
        config = super().get_config().copy()
        config.update({
            'filters': self.filters})
        return config

def LeNET():
    ip = tflayers.Input(shape=(32, 32, 3))
    c1 = customConv2d(6, 5, kernel_type, model_regularizer)(ip)
    c1 = tflayers.Activation('relu')(c1)
    cm1 = tflayers.MaxPooling2D((2, 2))(c1)
    c2 = customConv2d(16, 5, kernel_type, model_regularizer)(cm1) 
    c2 = tflayers.Activation('relu')(c2)
    cm2 = tflayers.MaxPooling2D((2, 2))(c2)
    c3 = customConv2d(120, 5, kernel_type, model_regularizer)(cm2)
    c3 = tflayers.Activation('relu')(c3)

    if topology=='2':
        cm2_2 = tflayers.MaxPooling2D((6, 6))(c2)
        c3 = tflayers.Concatenate()([c3, cm2_2])
        
    flatten = tflayers.Flatten()(c3)
    dense = tflayers.Dense(84, activation='relu')(flatten)
    op = tflayers.Dense(4, activation='softmax')(dense)

    model = tf.keras.Model(ip, op, name = 'LeNet5_'+topology)
    model.compile(optimizer=model_optimizer, loss=tf.keras.losses.SparseCategoricalCrossentropy(), metrics=['acc'])
    print("Finished creating model", 'LeNet5_'+topology)
    return model


data_folder = os.path.abspath(data_folder)
imgPaths = glob(data_folder+'/*')

X = []
Y = []

suit_encoder = {'C':0, 'D':1, 'H':2, 'S':3}

print("Importing the data from the folder", data_folder)

for imgPath in imgPaths:
    img = load_img(imgPath)
    img_array = img_to_array(img)/255
    suit = os.path.basename(imgPath).split('_')[1].split('.')[0]
    Y.append(suit_encoder[suit])
    X.append(img_array)

# X = tf.stack(X)
X = np.array(X)
Y = np.array(Y)

print('Finished importing')


if train_or_test == 'train':
    model = LeNET()
    start=time()
    model.fit(X,Y,epochs=epochs,batch_size=batch_size,verbose=1)
    end = time()-start
    print('Training data size:', len(Y))
    print("Time taken to train:", end)
    #print('\nTraining accuracy: ',max(history.history['val_accuracy']))
    model.save(model_file_name)

else:
    model=load_model(model_file_name,custom_objects={'customConv2d': customConv2d()})
    print('Testing started')
    start=time()
    predictions = model.predict(X).argmax(axis=-1)
    end = time()-start
    print('Testing complete\n')
    print('Testing data size:', len(Y))
    print("Time taken to test:", end)
    print("Accuracy of model:", (len(np.where(Y == predictions)[0])/len(Y))*100)
    print('Confusion Matrix:\n', tf.math.confusion_matrix(Y, predictions))